title: A practical introduction to functional programming
url: http://maryrosecook.com/blog/post/a-practical-introduction-to-functional-programming
hash_url: 437dce93077f9a2569566bfcfe47134d

<p>Many functional programming articles teach abstract functional techniques.  That is, composition, pipelining, higher order functions.  This one is different.  It shows examples of imperative, unfunctional code that people write every day and translates these examples to a functional style.</p>
<p>The first section of the article takes short, data transforming loops and translates them into functional maps and reduces.  The second section takes longer loops, breaks them up into units and makes each unit functional.  The third section takes a loop that is a long series of successive data transformations and decomposes it into a functional pipeline.</p>
<p>The examples are in Python, because many people find Python easy to read.  A number of the examples eschew pythonicity in order to demonstrate functional techniques common to many languages: map, reduce, pipeline.</p>
<h3>A guide rope</h3>
<p>When people talk about functional programming, they mention a dizzying number of “functional” characteristics.  They mention immutable data<sup>1</sup>, first class functions<sup>2</sup> and tail call optimisation<sup>3</sup>.  These are language features that aid functional programming.  They mention mapping, reducing, pipelining, recursing, currying<sup>4</sup> and the use of higher order functions.  These are programming techniques used to write functional code.  They mention parallelization<sup>5</sup>, lazy evaluation<sup>6</sup> and determinism<sup>7</sup>.  These are advantageous properties of functional programs.</p>
<p>Ignore all that.  Functional code is characterised by one thing: the absence of side effects.  It doesn’t rely on data outside the current function, and it doesn’t change data that exists outside the current function.  Every other “functional” thing can be derived from this property.  Use it as a guide rope as you learn.</p>
<p>This is an unfunctional function:</p>
<pre class="prettyprint">&#13;
a = 0&#13;
def increment1():&#13;
    global a&#13;
    a += 1&#13;
</pre>
<p>This is a functional function:</p>
<pre class="prettyprint">&#13;
def increment2(a):&#13;
    return a + 1&#13;
</pre>
<h3>Don’t iterate over lists.  Use map and reduce.</h3>
<h4>Map</h4>
<p>Map takes a function and a collection of items.  It makes a new, empty collection, runs the function on each item in the original collection and inserts each return value into the new collection.  It returns the new collection.</p>
<p>This is a simple map that takes a list of names and returns a list of the lengths of those names:</p>
<pre class="prettyprint">&#13;
name_lengths = map(len, ["Mary", "Isla", "Sam"])&#13;
&#13;
print name_lengths&#13;
# =&gt; [4, 4, 3]&#13;
</pre>
<p>This is a map that squares every number in the passed collection:</p>
<pre class="prettyprint">&#13;
squares = map(lambda x: x * x, [0, 1, 2, 3, 4])&#13;
&#13;
print squares&#13;
# =&gt; [0, 1, 4, 9, 16]&#13;
</pre>
<p>This map doesn’t take a named function.  It takes an anonymous, inlined function defined with <code>lambda</code>.  The parameters of the lambda are defined to the left of the colon.  The function body is defined to the right of the colon.  The result of running the function body is (implicitly) returned.</p>
<p>The unfunctional code below takes a list of real names and replaces them with randomly assigned code names.</p>
<pre class="prettyprint">&#13;
import random&#13;
&#13;
names = ['Mary', 'Isla', 'Sam']&#13;
code_names = ['Mr. Pink', 'Mr. Orange', 'Mr. Blonde']&#13;
&#13;
for i in range(len(names)):&#13;
    names[i] = random.choice(code_names)&#13;
&#13;
print names&#13;
# =&gt; ['Mr. Blonde', 'Mr. Blonde', 'Mr. Blonde']&#13;
</pre>
<p>(As you can see, this algorithm can potentially assign the same secret code name to multiple secret agents.  Hopefully, this won’t be a source of confusion during the secret mission.)</p>
<p>This can be rewritten as a map:</p>
<pre class="prettyprint">&#13;
import random&#13;
&#13;
names = ['Mary', 'Isla', 'Sam']&#13;
&#13;
secret_names = map(lambda x: random.choice(['Mr. Pink',&#13;
                                            'Mr. Orange',&#13;
                                            'Mr. Blonde']),&#13;
                   names)&#13;
</pre>
<p><strong>Exercise 1</strong>. Try rewriting the code below as a map.  It takes a list of real names and replaces them with code names produced using a more robust strategy.</p>
<pre class="prettyprint">&#13;
names = ['Mary', 'Isla', 'Sam']&#13;
&#13;
for i in range(len(names)):&#13;
    names[i] = hash(names[i])&#13;
&#13;
print names&#13;
# =&gt; [6306819796133686941, 8135353348168144921, -1228887169324443034]&#13;
</pre>
<p>(Hopefully, the secret agents will have good memories and won’t forget each other’s secret code names during the secret mission.)</p>
<p>My solution:</p>
<pre class="prettyprint">&#13;
names = ['Mary', 'Isla', 'Sam']&#13;
&#13;
secret_names = map(hash, names)&#13;
</pre>
<h4>Reduce</h4>
<p>Reduce takes a function and a collection of items.  It returns a value that is created by combining the items.</p>
<p>This is a simple reduce.  It returns the sum of all the items in the collection.</p>
<pre class="prettyprint">&#13;
sum = reduce(lambda a, x: a + x, [0, 1, 2, 3, 4])&#13;
&#13;
print sum&#13;
# =&gt; 10&#13;
</pre>
<p><code>x</code> is the current item being iterated over.  <code>a</code> is the accumulator.  It is the value returned by the execution of the lambda on the previous item.  <code>reduce()</code> walks through the items.  For each one, it runs the lambda on the current <code>a</code> and <code>x</code> and returns the result as the <code>a</code> of the next iteration.</p>
<p>What is <code>a</code> in the first iteration? There is no previous iteration result for it to pass along.  <code>reduce()</code> uses the first item in the collection for <code>a</code> in the first iteration and starts iterating at the second item.  That is, the first <code>x</code> is the second item.</p>
<p>This code counts how often the word <code>'Sam'</code> appears in a list of strings:</p>
<pre class="prettyprint">&#13;
sentences = ['Mary read a story to Sam and Isla.',&#13;
             'Isla cuddled Sam.',&#13;
             'Sam chortled.']&#13;
&#13;
sam_count = 0&#13;
for sentence in sentences:&#13;
    sam_count += sentence.count('Sam')&#13;
&#13;
print sam_count&#13;
# =&gt; 3&#13;
</pre>
<p>This is the same code written as a reduce:</p>
<pre class="prettyprint">&#13;
sentences = ['Mary read a story to Sam and Isla.',&#13;
             'Isla cuddled Sam.',&#13;
             'Sam chortled.']&#13;
&#13;
sam_count = reduce(lambda a, x: a + x.count('Sam'),&#13;
                   sentences,&#13;
                   0)&#13;
</pre>
<p>How does this code come up with its initial <code>a</code>? The starting point for the number of incidences of <code>'Sam'</code> cannot be <code>'Mary read a story to Sam and Isla.'</code>  The initial accumulator is specified with the third argument to <code>reduce()</code>.  This allows the use of a value of a different type from the items in the collection.</p>
<p>Why are map and reduce better?</p>
<p>First, they are often one-liners.</p>
<p>Second, the important parts of the iteration – the collection, the operation and the return value – are always in the same places in every map and reduce.</p>
<p>Third, the code in a loop may affect variables defined before it or code that runs after it.  By convention, maps and reduces are functional.</p>
<p>Fourth, map and reduce are elemental operations.  Every time a person reads a <code>for</code> loop, they have to work through the logic line by line.  There are few structural regularities they can use to create a scaffolding on which to hang their understanding of the code.  In contrast, map and reduce are at once building blocks that can be combined into complex algorithms, and elements that the code reader can instantly understand and abstract in their mind.  “Ah, this code is transforming each item in this collection.  It’s throwing some of the transformations away.  It’s combining the remainder into a single output.”</p>
<p>Fifth, map and reduce have many friends that provide useful, tweaked versions of their basic behaviour.  For example: <code>filter</code>, <code>all</code>, <code>any</code> and <code>find</code>.</p>
<p><strong>Exercise 2</strong>.  Try rewriting the code below using map, reduce and filter.  Filter takes a function and a collection.  It returns a collection of every item for which the function returned <code>True</code>.</p>
<pre class="prettyprint">&#13;
people = [{'name': 'Mary', 'height': 160},&#13;
          {'name': 'Isla', 'height': 80},&#13;
          {'name': 'Sam'}]&#13;
&#13;
height_total = 0&#13;
height_count = 0&#13;
for person in people:&#13;
    if 'height' in person:&#13;
        height_total += person['height']&#13;
        height_count += 1&#13;
&#13;
if height_count &gt; 0:&#13;
    average_height = height_total / height_count&#13;
&#13;
    print average_height&#13;
    # =&gt; 120&#13;
</pre>
<p>If this seems tricky, try not thinking about the operations on the data.  Think of the states the data will go through, from the list of people dictionaries to the average height.  Don’t try and bundle multiple transformations together.  Put each on a separate line and assign the result to a descriptively-named variable.  Once the code works, condense it.</p>
<p>My solution:</p>
<pre class="prettyprint">&#13;
people = [{'name': 'Mary', 'height': 160},&#13;
          {'name': 'Isla', 'height': 80},&#13;
          {'name': 'Sam'}]&#13;
&#13;
heights = map(lambda x: x['height'],&#13;
              filter(lambda x: 'height' in x, people))&#13;
&#13;
if len(heights) &gt; 0:&#13;
    from operator import add&#13;
    average_height = reduce(add, heights) / len(heights)&#13;
</pre>
<h3>Write declaratively, not imperatively</h3>
<p>The program below runs a race between three cars.  At each time step, each car may move forwards or it may stall.  At each time step, the program prints out the paths of the cars so far.  After five time steps, the race is over.</p>
<p>This is some sample output:</p>
<pre class="prettyprint">&#13;
-&#13;
--&#13;
--&#13;
&#13;
--&#13;
--&#13;
---&#13;
&#13;
---&#13;
--&#13;
---&#13;
&#13;
----&#13;
---&#13;
----&#13;
&#13;
----&#13;
----&#13;
-----&#13;
</pre>
<p>This is the program:</p>
<pre class="prettyprint">&#13;
from random import random&#13;
&#13;
time = 5&#13;
car_positions = [1, 1, 1]&#13;
&#13;
while time:&#13;
    # decrease time&#13;
    time -= 1&#13;
&#13;
    print ''&#13;
    for i in range(len(car_positions)):&#13;
        # move car&#13;
        if random() &gt; 0.3:&#13;
            car_positions[i] += 1&#13;
&#13;
        # draw car&#13;
        print '-' * car_positions[i]&#13;
</pre>
<p>The code is written imperatively.  A functional version would be declarative.  It would describe what to do, rather than how to do it.</p>
<h4>Use functions</h4>
<p>A program can be made more declarative by bundling pieces of the code into functions.</p>
<pre class="prettyprint">&#13;
from random import random&#13;
&#13;
def move_cars():&#13;
    for i, _ in enumerate(car_positions):&#13;
        if random() &gt; 0.3:&#13;
            car_positions[i] += 1&#13;
&#13;
def draw_car(car_position):&#13;
    print '-' * car_position&#13;
&#13;
def run_step_of_race():&#13;
    global time&#13;
    time -= 1&#13;
    move_cars()&#13;
&#13;
def draw():&#13;
    print ''&#13;
    for car_position in car_positions:&#13;
        draw_car(car_position)&#13;
&#13;
time = 5&#13;
car_positions = [1, 1, 1]&#13;
&#13;
while time:&#13;
    run_step_of_race()&#13;
    draw()&#13;
</pre>
<p>To understand this program, the reader just reads the main loop.  “If there is time left, run a step of the race and draw.  Check the time again.”  If the reader wants to understand more about what it means to run a step of the race, or draw, they can read the code in those functions.</p>
<p>There are no comments any more.  The code describes itself.</p>
<p>Splitting code into functions is a great, low brain power way to make code more readable.</p>
<p>This technique uses functions, but it uses them as sub-routines. They parcel up code.  The code is not functional in the sense of the guide rope.  The functions in the code use state that was not passed as arguments.  They affect the code around them by changing external variables, rather than by returning values.  To check what a function really does, the reader must read each line carefully.  If they find an external variable, they must find its origin.  They must see what other functions change that variable.</p>
<h4>Remove state</h4>
<p>This is a functional version of the car race code:</p>
<pre class="prettyprint">&#13;
from random import random&#13;
&#13;
def move_cars(car_positions):&#13;
    return map(lambda x: x + 1 if random() &gt; 0.3 else x,&#13;
               car_positions)&#13;
&#13;
def output_car(car_position):&#13;
    return '-' * car_position&#13;
&#13;
def run_step_of_race(state):&#13;
    return {'time': state['time'] - 1,&#13;
            'car_positions': move_cars(state['car_positions'])}&#13;
&#13;
def draw(state):&#13;
    print ''&#13;
    print '\n'.join(map(output_car, state['car_positions']))&#13;
&#13;
def race(state):&#13;
    draw(state)&#13;
    if state['time']:&#13;
        race(run_step_of_race(state))&#13;
&#13;
race({'time': 5,&#13;
      'car_positions': [1, 1, 1]})&#13;
</pre>
<p>The code is still split into functions, but the functions are functional.  There are three signs of this.  First, there are no longer any shared variables.  <code>time</code> and <code>car_positions</code> get passed straight into <code>race()</code>.  Second, functions take parameters.  Third, no variables are instantiated inside functions.  All data changes are done with return values.  <code>race()</code> recurses<sup>3</sup> with the result of <code>run_step_of_race()</code>.  Each time a step generates a new state, it is passed immediately into the next step.</p>
<p>Now, here are two functions, <code>zero()</code> and <code>one()</code>:</p>
<pre class="prettyprint">&#13;
def zero(s):&#13;
    if s[0] == "0":&#13;
        return s[1:]&#13;
&#13;
def one(s):&#13;
    if s[0] == "1":&#13;
        return s[1:]&#13;
</pre>
<p><code>zero()</code> takes a string, <code>s</code>.  If the first character is <code>'0'</code>, it returns the rest of the string.  If it is not, it returns <code>None</code>, the default return value of Python functions.  <code>one()</code> does the same, but for a first character of <code>'1'</code>.</p>
<p>Imagine a function called <code>rule_sequence()</code>.  It takes a string and a list of rule functions of the form of <code>zero()</code> and <code>one()</code>.  It calls the first rule on the string.  Unless <code>None</code> is returned, it takes the return value and calls the second rule on it.  Unless <code>None</code> is returned, it takes the return value and calls the third rule on it.  And so forth.  If any rule returns <code>None</code>, <code>rule_sequence()</code> stops and returns <code>None</code>.  Otherwise, it returns the return value of the final rule.</p>
<p>This is some sample input and output:</p>
<pre class="prettyprint">&#13;
print rule_sequence('0101', [zero, one, zero])&#13;
# =&gt; 1&#13;
&#13;
print rule_sequence('0101', [zero, zero])&#13;
# =&gt; None&#13;
</pre>
<p>This is the imperative version of <code>rule_sequence()</code>:</p>
<pre class="prettyprint">&#13;
def rule_sequence(s, rules):&#13;
    for rule in rules:&#13;
        s = rule(s)&#13;
        if s == None:&#13;
            break&#13;
&#13;
    return s&#13;
</pre>
<p><strong>Exercise 3</strong>.  The code above uses a loop to do its work.   Make it more declarative by rewriting it as a recursion.</p>
<p>My solution:</p>
<pre class="prettyprint">&#13;
def rule_sequence(s, rules):&#13;
    if s == None or not rules:&#13;
        return s&#13;
    else:&#13;
        return rule_sequence(rules[0](s), rules[1:])&#13;
</pre>
<h3>Use pipelines</h3>
<p>In the previous section, some imperative loops were rewritten as recursions that called out to auxiliary functions.  In this section, a different type of imperative loop will be rewritten using a technique called a pipeline.</p>
<p>The loop below performs transformations on dictionaries that hold the name, incorrect country of origin and active status of some bands.</p>
<pre class="prettyprint">&#13;
bands = [{'name': 'sunset rubdown', 'country': 'UK', 'active': False},&#13;
         {'name': 'women', 'country': 'Germany', 'active': False},&#13;
         {'name': 'a silver mt. zion', 'country': 'Spain', 'active': True}]&#13;
&#13;
def format_bands(bands):&#13;
    for band in bands:&#13;
        band['country'] = 'Canada'&#13;
        band['name'] = band['name'].replace('.', '')&#13;
        band['name'] = band['name'].title()&#13;
&#13;
format_bands(bands)&#13;
&#13;
print bands&#13;
# =&gt; [{'name': 'Sunset Rubdown', 'active': False, 'country': 'Canada'},&#13;
#     {'name': 'Women', 'active': False, 'country': 'Canada' },&#13;
#     {'name': 'A Silver Mt Zion', 'active': True, 'country': 'Canada'}]&#13;
</pre>
<p>Worries are stirred by the name of the function.  “format” is very vague.  Upon closer inspection of the code, these worries begin to claw.  Three things happen in the same loop.  The <code>'country'</code> key gets set to <code>'Canada'</code>.  Punctuation is removed from the band name.  The band name gets capitalized.  It is hard to tell what the code is intended to do and hard to tell if it does what it appears to do.  The code is hard to reuse, hard to test and hard to parallelize.</p>
<p>Compare it with this:</p>
<pre class="prettyprint">&#13;
print pipeline_each(bands, [set_canada_as_country,&#13;
                            strip_punctuation_from_name,&#13;
                            capitalize_names])&#13;
</pre>
<p>This code is easy to understand.  It gives the impression that the auxiliary functions are functional because they seem to be chained together.  The output from the previous one comprises the input to the next.  If they are functional, they are easy to verify.  They are also easy to reuse, easy to test and easy to parallelize.</p>
<p>The job of <code>pipeline_each()</code> is to pass the bands, one at a time, to a transformation function, like <code>set_canada_as_country()</code>.  After the function has been applied to all the bands, <code>pipeline_each()</code> bundles up the transformed bands.  Then, it passes each one to the next function.</p>
<p>Let’s look at the transformation functions.</p>
<pre class="prettyprint">&#13;
def assoc(_d, key, value):&#13;
    from copy import deepcopy&#13;
    d = deepcopy(_d)&#13;
    d[key] = value&#13;
    return d&#13;
&#13;
def set_canada_as_country(band):&#13;
    return assoc(band, 'country', "Canada")&#13;
&#13;
def strip_punctuation_from_name(band):&#13;
    return assoc(band, 'name', band['name'].replace('.', ''))&#13;
&#13;
def capitalize_names(band):&#13;
    return assoc(band, 'name', band['name'].title())&#13;
</pre>
<p>Each one associates a key on a band with a new value.  There is no easy way to do this without mutating the original band.  <code>assoc()</code> solves this problem by using <code>deepcopy()</code> to produce a copy of the passed dictionary.  Each transformation function makes its modification to the copy and returns that copy.</p>
<p>Everything seems fine.  Band dictionary originals are protected from mutation when a key is associated with a new value.  But there are two other potential mutations in the code above.  In <code>strip_punctuation_from_name()</code>, the unpunctuated name is generated by calling <code>replace()</code> on the original name.  In <code>capitalize_names()</code>, the capitalized name is generated by calling <code>title()</code> on the original name.  If <code>replace()</code> and <code>title()</code> are not functional,  <code>strip_punctuation_from_name()</code> and <code>capitalize_names()</code> are not functional.</p>
<p>Fortunately, <code>replace()</code> and <code>title()</code> do not mutate the strings they operate on.  This is because strings are immutable in Python.  When, for example, <code>replace()</code> operates on a band name string, the original band name is copied and <code>replace()</code> is called on the copy.  Phew.</p>
<p>This contrast between the mutability of strings and dictionaries in Python illustrates the appeal of languages like Clojure.  The programmer need never think about whether they are mutating data.  They aren’t.</p>
<p><strong>Exercise 4</strong>.  Try and write the <code>pipeline_each</code> function.  Think about the order of operations.  The bands in the array are passed, one band at a time, to the first transformation function.  The bands in the resulting array are passed, one band at a time, to the second transformation function.  And so forth.</p>
<p>My solution:</p>
<pre class="prettyprint">&#13;
def pipeline_each(data, fns):&#13;
    return reduce(lambda a, x: map(x, a),&#13;
                  fns,&#13;
                  data)&#13;
</pre>
<p>All three transformation functions boil down to making a change to a particular field on the passed band.  <code>call()</code> can be used to abstract that.  It takes a function to apply and the key of the value to apply it to.</p>
<pre class="prettyprint">&#13;
set_canada_as_country = call(lambda x: 'Canada', 'country')&#13;
strip_punctuation_from_name = call(lambda x: x.replace('.', ''), 'name')&#13;
capitalize_names = call(str.title, 'name')&#13;
&#13;
print pipeline_each(bands, [set_canada_as_country,&#13;
                            strip_punctuation_from_name,&#13;
                            capitalize_names])&#13;
</pre>
<p>Or, if we are willing to sacrifice readability for conciseness, just:</p>
<pre class="prettyprint">&#13;
print pipeline_each(bands, [call(lambda x: 'Canada', 'country'),&#13;
                            call(lambda x: x.replace('.', ''), 'name'),&#13;
                            call(str.title, 'name')])&#13;
</pre>
<p>The code for <code>call()</code>:</p>
<pre class="prettyprint">&#13;
def assoc(_d, key, value):&#13;
    from copy import deepcopy&#13;
    d = deepcopy(_d)&#13;
    d[key] = value&#13;
    return d&#13;
&#13;
def call(fn, key):&#13;
    def apply_fn(record):&#13;
        return assoc(record, key, fn(record.get(key)))&#13;
    return apply_fn&#13;
</pre>
<p>There is a lot going on here.  Let’s take it piece by piece.</p>
<p>One.  <code>call()</code> is a higher order function.  A higher order function takes a function as an argument, or returns a function.  Or, like <code>call()</code>, it does both.</p>
<p>Two.  <code>apply_fn()</code> looks very similar to the three transformation functions.  It takes a record (a band).  It looks up the value at <code>record[key]</code>.  It calls <code>fn</code> on that value.  It assigns the result back to a copy of the record.  It returns the copy.</p>
<p>Three.  <code>call()</code> does not do any actual work.  <code>apply_fn()</code>, when called, will do the work.  In the example of using <code>pipeline_each()</code> above, one instance of <code>apply_fn()</code> will set <code>'country'</code> to <code>'Canada'</code> on a passed band.  Another instance will capitalize the name of a passed band.</p>
<p>Four.  When an <code>apply_fn()</code> instance is run, <code>fn</code> and <code>key</code> will not be in scope.  They are neither arguments to <code>apply_fn()</code>, nor locals inside it.  But they will still be accessible.  When a function is defined, it saves references to the variables it closes over: those that were defined in a scope outside the function and that are used inside the function.  When the function is run and its code references a variable, Python looks up the variable in the locals and in the arguments.  If it doesn’t find it there, it looks in the saved references to closed over variables.  This is where it will find <code>fn</code> and <code>key</code>.</p>
<p>Five.  There is no mention of bands in the <code>call()</code> code.  That is because <code>call()</code> could be used to generate pipeline functions for any program, regardless of topic.  Functional programming is partly about building up a library of generic, reusable, composable functions.</p>
<p>Good job.  Closures, higher order functions and variable scope all covered in the space of a few paragraphs.  Have a nice glass of lemonade.</p>
<p>There is one more piece of band processing to do.  That is to remove everything but the name and country.  <code>extract_name_and_country()</code> can pull that information out:</p>
<pre class="prettyprint">&#13;
def extract_name_and_country(band):&#13;
    plucked_band = {}&#13;
    plucked_band['name'] = band['name']&#13;
    plucked_band['country'] = band['country']&#13;
    return plucked_band&#13;
&#13;
print pipeline_each(bands, [call(lambda x: 'Canada', 'country'),&#13;
                            call(lambda x: x.replace('.', ''), 'name'),&#13;
                            call(str.title, 'name'),&#13;
                            extract_name_and_country])&#13;
&#13;
# =&gt; [{'name': 'Sunset Rubdown', 'country': 'Canada'},&#13;
#     {'name': 'Women', 'country': 'Canada'},&#13;
#     {'name': 'A Silver Mt Zion', 'country': 'Canada'}]&#13;
</pre>
<p><code>extract_name_and_country()</code> could have been written as a generic function called <code>pluck()</code>.  <code>pluck()</code> would be used like this:</p>
<pre class="prettyprint">&#13;
print pipeline_each(bands, [call(lambda x: 'Canada', 'country'),&#13;
                            call(lambda x: x.replace('.', ''), 'name'),&#13;
                            call(str.title, 'name'),&#13;
                            pluck(['name', 'country'])])&#13;
</pre>
<p><strong>Exercise 5</strong>. <code>pluck()</code> takes a list of keys to extract from each record.  Try and write it.  It will need to be a higher order function.</p>
<p>My solution:</p>
<pre class="prettyprint">&#13;
def pluck(keys):&#13;
    def pluck_fn(record):&#13;
        return reduce(lambda a, x: assoc(a, x, record[x]),&#13;
                      keys,&#13;
                      {})&#13;
    return pluck_fn&#13;
</pre>
<h3>What now?</h3>
<p>Functional code co-exists very well with code written in other styles.  The transformations in this article can be applied to any code base in any language.  Try applying them to your own code.</p>
<p>Think of Mary, Isla and Sam.  Turn iterations of lists into maps and reduces.</p>
<p>Think of the race.  Break code into functions.  Make those functions functional.  Turn a loop that repeats a process into a recursion.</p>
<p>Think of the bands.  Turn a sequence of operations into a pipeline.</p>
<p><br/></p>
<p><sup>1</sup> An immutable piece of data is one that cannot be changed.  Some languages, like Clojure, make all values immutable by default.  Any “mutating” operations copy the value, change it and pass back the changed copy.  This eliminates bugs that arise from a programmer’s incomplete model of the possible states their program may enter.</p>
<p><sup>2</sup> Languages that support first class functions allow functions to be treated like any other value.  This means they can be created, passed to functions, returned from functions and stored inside data structures.</p>
<p><sup>3</sup> Tail call optimisation is a programming language feature.  Each time a function recurses, a new stack frame is created.  A stack frame is used to store the arguments and local values for the current function invocation.  If a function recurses a large number of times, it is possible for the interpreter or compiler to run out of memory.  Languages with tail call optimisation reuse the same stack frame for their entire sequence of recursive calls.  Languages like Python that do not have tail call optimisation generally limit the number of times a function may recurse to some number in the thousands.  In the case of the <code>race()</code> function, there are only five time steps, so it is safe.</p>
<p><sup>4</sup> Currying means decomposing a function that takes multiple arguments into a function that takes the first argument and returns a function that takes the next argument, and so forth for all the arguments.</p>
<p><sup>5</sup> Parallelization means running the same code concurrently without synchronization.  These concurrent processes are often run on multiple processors.</p>
<p><sup>6</sup> Lazy evaluation is a compiler technique that avoids running code until the result is needed.</p>
<p><sup>7</sup> A process is deterministic if repetitions yield the same result every time.</p>
